from examples.feature_metrics import main as feature_metrics_examples
from examples.image_metrics import main as image_metrics_examples


def test_image_metrics():
    image_metrics_examples()


def test_feature_metrics():
    feature_metrics_examples()
